
from ... conditional import ordereddict


class namespace_cls:
    """
    similar to types.SimpleNamespace
    """
    def __init__(self, **kwargs):
        self.__dict__ = ordereddict(kwargs)

    def __iter__(self):
        return ((k, v) for k, v in self.__dict__.items())

    def __eq__(self, other):
        if (not hasattr(self, '__dict__')) or (not hasattr(other, '__dict__')):
            return NotImplemented

        return self.__dict__ == other.__dict__

    def __repr__(self):
        items = ('{}={!r}'.format(k, v) for k, v in self.__dict__.items())
        items = ', '.join(items)

        return '{{{}}}'.format(items)





import csv
import gc
import os
import pickle
import pprint
import shutil

from collections import namedtuple
from datetime import date
from datetime import datetime
from glob import glob
from io import StringIO
from os.path import isdir  as os_isdir
from os.path import isfile as os_isfile
from urllib.parse import urlparse
from urllib.request import urlopen

from ..conditional import is_windows_os
from ..conditional import ultrajson_installed

if ultrajson_installed:
    # noinspection PyUnresolvedReferences
    import ujson as json
else:
    import json

pickle_extensions = {'.flux',
                     '.pkl',
                     '.pickle'}


def read_file(path,
              encoding=None,
              mode='r',
              filetype=None,
              **kwargs):

    path, encoding, filetype, mode, kwargs = \
     __validate_io_arguments(path, encoding, mode, filetype, kwargs, 'read')

    is_bytes_mode  = ('b' in mode)
    is_url         = is_path_a_url(path)
    was_gc_enabled = gc.isenabled()

    gc.disable()

    if filetype == '.csv':
        data = __read_csv(path, mode, encoding, kwargs)

    elif filetype == '.json':
        data = __read_json(path, mode, encoding, kwargs)

    elif filetype in pickle_extensions:
        with open(path, mode) as f:
            data = pickle.load(f, **kwargs)

    elif is_url:
        data = __url_request(path, encoding=encoding)

    elif is_bytes_mode:
        with open(path, mode) as f:
            data = f.read()

    else:
        with open(path, mode, encoding=encoding) as f:
            data = f.read()

    if was_gc_enabled:
        gc.enable()

    return data


# noinspection DuplicatedCode
def write_file(path,
               data,
               encoding=None,
               mode='w',
               filetype=None,
               **kwargs):

    path, encoding, filetype, mode, kwargs = \
     __validate_io_arguments(path, encoding, mode, filetype, kwargs, 'write')

    is_bytes_mode  = ('b' in mode)
    was_gc_enabled = gc.isenabled()

    gc.disable()

    if filetype == '.csv':
        newline = kwargs.pop('newline')
        with open(path, mode, encoding=encoding, newline=newline) as f:
            csv.writer(f, **kwargs).writerows(data)

    elif filetype == '.json':
        with open(path, mode, encoding=encoding) as f:
            json.dump(data, f, **kwargs)

    elif filetype in pickle_extensions:
        with open(path, mode) as f:
            pickle.dump(data, f, **kwargs)

    elif is_bytes_mode:
        with open(path, mode) as f:
            f.write(data)

    else:
        if not isinstance(data, str): data = pprint.pformat(data, **kwargs)
        with open(path, mode, encoding=encoding) as f:
            f.write(data)

    if was_gc_enabled:
        gc.enable()


def __validate_io_arguments(path,
                            encoding,
                            mode,
                            filetype,
                            kwargs,
                            read_or_write):

    path          = __validate_path(path)
    filetype      = __validate_filetype(path, filetype)
    mode          = __validate_mode(mode, filetype)
    # encoding      = __validate_encoding(encoding, mode, filetype)
    read_or_write = __validate_read_or_write(path, mode, read_or_write)
    kwargs        = __validate_file_keyword_args(encoding, filetype, kwargs, read_or_write)

    return (path,
            encoding,
            filetype,
            mode,
            kwargs)


def __validate_path(path):
    path = str(path)
    path = standardize_path(path, explicit_cwd=True)

    return path


def __validate_filetype(path, filetype):
    """ check for file types that require specialized io libraries / protocols """
    notimplemented_extensions = {'.7z',
                                 '.zip',
                                 '.gzip',
                                 '.bz',
                                 '.tar',
                                 '.png',
                                 '.jpg',
                                 '.jpeg',
                                 '.gif',
                                 '.pdf'}

    filetype = filetype or parse_file_extension(path, include_dot=True)
    filetype = filetype.lower().strip()

    if not filetype:
        raise TypeError('empty file type')
    elif not filetype.startswith('.'):
        filetype = '.' + filetype

    if filetype.startswith('.xl') or filetype in notimplemented_extensions:
        raise NotImplementedError("file type not supported: '{}'".format(filetype))

    return filetype


def __validate_mode(mode, filetype):
    is_rw_mode          = ('+' in mode)
    is_bytes_mode       = ('b' in mode)
    is_pickle_extension = (filetype in pickle_extensions)
    is_csv_extension    = (filetype == '.csv')

    if is_rw_mode:
        raise ValueError('read-write mode not supported')
    if is_csv_extension and is_bytes_mode:
        raise ValueError('csv filetype does not accept bytes mode')

    if is_pickle_extension and (not is_bytes_mode):
        mode = mode + 'b'

    return mode


# def __validate_encoding(encoding, mode, filetype):
#     is_bytes_mode = ('b' in mode)
#
#     if is_bytes_mode:
#         encoding = None
#
#     return encoding


def __validate_file_keyword_args(encoding, filetype, kwargs, read_or_write):
    kwargs = kwargs or {}

    if 'kwargs' in kwargs:
        kwargs.update(kwargs.pop('kwargs'))

    if read_or_write == 'read':
        if filetype == '.csv':
            kwargs['newline'] = kwargs.get('newline', '')
            kwargs['nrows']   = kwargs.get('nrows',   None)

    if read_or_write == 'write':
        if filetype == '.csv':
            kwargs['newline'] = kwargs.get('newline', '')

        if filetype == '.json':
            kwargs['indent']       = kwargs.get('indent',       4)
            kwargs['ensure_ascii'] = kwargs.get('ensure_ascii', encoding in (None, 'ascii'))
            kwargs['default']      = kwargs.get('default',      json_unhandled_conversion)

            if ultrajson_installed:
                for k in ('default', 'separators'):
                    try:             del kwargs[k]
                    except KeyError: pass

    return kwargs


# noinspection PyRedundantParentheses
def __validate_read_or_write(path, mode, read_or_write):
    is_read_mode   = ('r' in mode)
    is_write_mode  = ('w' in mode)
    is_append_mode = ('a' in mode)
    is_url         = is_path_a_url(path)

    if read_or_write == 'read':
        if (not is_read_mode) or is_append_mode:
            raise ValueError('invalid read mode: {}'.format(mode))

    elif read_or_write == 'write':
        if (not is_write_mode):
            raise ValueError('invalid write mode: {}'.format(mode))
        if is_url:
            raise NotImplementedError('can not write to url: {}'.format(path))

    return read_or_write


def __read_csv(path, mode, encoding, kwargs):
    """
    _csv.Error: new-line character seen in unquoted field - do you need to open the file in universal-newline mode?
        fixed by passing lineterminator='\r'
        or newline='\r', newline=''
    """
    # region {closure}
    def read_csv_rows():
        if read_all_rows:
            return list(csv_reader)

        m = []
        try:
            for i in range(nrows):
                m.append(next(csv_reader))
        except StopIteration:
            pass

        return m

    # noinspection PyUnusedLocal
    def remove_url_from_rows(m):
        num_deleted = 0

        try:
            while is_path_a_url(m[0][0]):
                del m[0]
                num_deleted += 1
        except Exception as e:
            pass

        # if nrows specified, append rows to make up for deleted rows
        if not read_all_rows:
            try:
                for _ in range(num_deleted):
                    m.append(next(csv_reader))
            except StopIteration:
                pass

        return m
    # endregion

    newline  = kwargs.pop('newline')
    nrows    = kwargs.pop('nrows')
    read_all_rows = (nrows is None)

    if is_path_a_url(path):
        with StringIO(__url_request(path, encoding=encoding), newline=newline) as f:
            csv_reader = csv.reader(f, **kwargs)
            csv_m = read_csv_rows()
            csv_m = remove_url_from_rows(csv_m)

            return csv_m

    with open(path, mode, encoding=encoding, newline=newline) as f:
        csv_reader = csv.reader(f, **kwargs)
        csv_m = read_csv_rows()

        return csv_m


def __read_json(path, mode, encoding, kwargs):
    if is_path_a_url(path):
        s = __url_request(path, encoding=encoding)
        return json.loads(s, **kwargs)

    with open(path, mode, encoding=encoding) as f:
        return json.load(f, **kwargs)


def __url_request(url, encoding=None):
    with urlopen(url) as request:
        if request.code != 200:
            raise IOError('bad url request: ({}) {}'.format(request.code, url))

        byte_string = request.read()

        if encoding:
            return byte_string.decode(encoding)
        else:
            return byte_string


def json_dumps_extended(o, **kwargs):
    kwargs['ensure_ascii'] = kwargs.get('ensure_ascii', False)
    kwargs['indent']       = kwargs.get('indent', 4)
    kwargs['default']      = kwargs.get('default', json_unhandled_conversion)

    if ultrajson_installed:
        del kwargs['default']

    return json.dumps(o, **kwargs)


def json_unhandled_conversion(v):
    """
    convert certain python objects to json string representations, eg:
        date, datetime, set
    """
    if isinstance(v, date):
        return v.isoformat()
    if isinstance(v, set):
        return list(v)
    # if isinstance(v, Decimal) ?

    raise TypeError("object '{}', of type '{}' is not JSON serializable".format(v, type(v)))


def clear_dir(filedir):
    filedir = standardize_dir(filedir, explicit_cwd=True)

    if not os.path.exists(filedir):
        return

    if not os_isdir(filedir):
        raise TypeError('"{}" is not a directory'.format(filedir))

    for item in os.listdir(filedir):
        path = filedir + item

        if os_isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


def copy_dir(s_dir, d_dir, ignore=None, dirs_exist_ok=False):
    """
    shutil.copytree will fail if shutil.rmtree hasn't finished deleting folder:
        PermissionError: [WinError 5] Access is denied
    """
    s_dir = standardize_dir(s_dir)
    d_dir = standardize_dir(d_dir)

    if ignore:
        if isinstance(ignore, str):
            ignore = (ignore,)
        elif isinstance(ignore, set):
            ignore = tuple(ignore)

        ignore = shutil.ignore_patterns(*ignore)

    if not dirs_exist_ok and os.path.exists(d_dir):
        shutil.rmtree(d_dir)

    shutil.copytree(src=s_dir,
                    dst=d_dir,
                    ignore=ignore,
                    dirs_exist_ok=dirs_exist_ok)


def file_creation_datetime(path):
    unix_t = os.path.getctime(path)
    return datetime.fromtimestamp(unix_t)


def file_modified_datetime(path):
    unix_t = os.stat(path).st_mtime
    return datetime.fromtimestamp(unix_t)


# @lru_cache(2**16)
def is_path_a_url(path):
    u = urlparse(str(path)).netloc
    return bool(u)


def parse_path(path,
               pathsep='/',
               explicit_cwd=False):

    ParsedPath = namedtuple('ParsedPath', ('directory',
                                           'filename',
                                           'extension'))

    _path_ = path
    _path_ = standardize_path(_path_, pathsep, explicit_cwd)

    filedir,  filename  = os.path.split(_path_)
    filename, extension = os.path.splitext(filename)

    return ParsedPath(filedir, filename, extension)


def parse_file_name(path):
    filename, _ = os.path.splitext(str(path))

    return filename


def parse_file_extension(filename, include_dot=True):
    _, extension = os.path.splitext(str(filename))
    if (include_dot is False) and extension.startswith('.'):
        extension = extension[1:]

    return extension


def standardize_path(path,
                     pathsep='/',
                     explicit_cwd=False):

    _path_ = path

    if not _path_:
        return ''

    if (not explicit_cwd) and (not os.path.isabs(path)):
        _path_ = os.path.relpath(_path_)
    else:
        _path_ = os.path.realpath(_path_)

    if pathsep == '/':
        _path_ = _path_.replace('\\', '/')
    else:
        _path_ = _path_.replace('/', '\\')

    if os_isdir(_path_):
        if _path_ and (not _path_.endswith(pathsep)):
            _path_ += pathsep

    return _path_


def standardize_dir(filedir,
                    pathsep='/',
                    explicit_cwd=False):

    return standardize_path(filedir, pathsep, explicit_cwd)


def traverse_dir(rootdir='.',
                 pathsep='/',
                 explicit_cwd=True,
                 *,
                 recurse=False,
                 subdirs_only=False,
                 files_only=False):

    if not is_windows_os:
        raise NotImplementedError('only available on windows OS')

    if (subdirs_only is False) and (files_only is False):
        subdirs_only = True
        files_only   = True

    rootdir = standardize_dir(rootdir, pathsep, explicit_cwd)

    if recurse:
        args = 'dir /b /s "{}"'.format(rootdir)
    else:
        args = 'dir /b "{}"'.format(rootdir)

    with os.popen(args) as os_cmd:
        out = os_cmd.read()
        out = out.split('\n')
        out.sort()

    if out and out[0] == '':
        del out[0]

    if recurse is False:
        out = [rootdir + p for p in out]

    if subdirs_only and files_only:
        return out
    elif subdirs_only:
        return [path for path in out
                     if os_isdir(path)]
    elif files_only:
        return [path for path in out
                     if os_isfile(path)]


def validate_path_exists(path):

    (filedir,
     filename,
     extension) = parse_path(path, explicit_cwd=True)

    path = filedir + filename + extension

    if not os.path.exists(filedir):
        raise FileNotFoundError('directory not found: \n\t{}'.format(filedir))

    if not os.path.exists(path):
        glob_paths = glob(filedir + filename + '.*')
        if glob_paths:

            extension = parse_file_extension(os.path.split(glob_paths[0])[1], include_dot=True)
            raise FileNotFoundError('file extension not found: '
                                    '\n\t{}'
                                    '\n\t{}'
                                    '\n\tdid you mean: {}?'.format(filedir, filename, extension))

        raise FileNotFoundError('file not found: '
                                '\n\t{}'
                                '\n\t{}'.format(filedir, filename + extension))

    return path




from collections import namedtuple

from typing import Dict
from typing import List
from typing import Union

from ..util.classes.namespace_cls import namespace_cls
from ..util.iter import is_header_row

from ..util.text import format_header
from ..util.text import format_header_lite
from ..util.text import format_integer

from ..conditional import ordereddict
from ..conditional import numpy_installed

if numpy_installed:
    import numpy


class flux_row_cls:

    @classmethod
    def reserved_names(cls):
        return ['_headers', 'values'] + dir(cls)

    def __init__(self, headers, values):
        """
        :param headers: OrderedDict of {'header': int}
            headers is a single dictionary passed byref from the flux_cls to many flux_row_cls instances
            this eliminates need for all flux_row_cls objects to maintain a seprate copy of these mappings and
            allows for centralized and instantaneous updatdes
        :param values: list of underlying data

        properties must be set on self.__dict__ instead of directly on self to prevent
        premature __setattr__ lookups
        """
        self._headers: Dict[Union[str, bytes], int]
        self.values:   List

        self.__dict__['_headers'] = headers
        self.__dict__['values']   = values

    @property
    def as_array(self):
        """
        to help with debugging: meant to trigger a debugging feature in PyCharm
        PyCharm will recognize the ndarray and enable the "...view as array"
        option in the debugger which displays values in a special window as a table
        """
        names  = [format_header(n) for n in self.header_names()]
        values = list(self.__dict__['values'])

        c_m = max(len(names), len(values))
        names.extend(['🗲']          * (c_m - len(names)))
        values.extend(['🗲jagged🗲'] * (c_m - len(values)))

        if 'address' in self.__dict__:
            names.insert(0,  '⟨address⟩')
            values.insert(0, '⟨{}⟩'.format(self.__dict__['address']))
        elif 'r_i' in self.__dict__:
            names.insert(0,  '⟨address⟩')
            values.insert(0, '⟨{:,}⟩'.format(self.__dict__['r_i']).replace(',', '_'))

        m = list(zip(names, values))

        if not numpy_installed:
            _nf_ = max([len(n) for n in names])  + 1
            _nf_ = '{: <%s}' % str(_nf_)

            m = [' '.join([_nf_.format(n), v]) for n, v in m]
            m = '\n'.join(m)

            return m

        return numpy.array(m, dtype=object)

    @property
    def headers(self):
        return ordereddict(self.__dict__['_headers'].items())

    def header_names(self):
        return list(self.__dict__['_headers'].keys())

    def is_jagged(self):
        return len(self.__dict__['_headers']) != len(self.__dict__['values'])

    def is_empty(self):
        return (len(self.__dict__['_headers']) == 0) and (len(self.__dict__['values']) == 0)

    def is_header_row(self):
        """ determine if underlying values match self._headers.keys

        self.names == self.values will not always work, since map_values_to_enum()
        was used to modify self._headers values into more suitable dictionary keys,
        like modifying duplicate values to ensure they are unique, etc
        """
        return is_header_row(self.__dict__['values'],
                             self.__dict__['_headers'])

    def dict(self):
        return ordereddict(zip(self.header_names(), self.__dict__['values']))

    def namedrow(self):
        d = ordereddict(zip(self.header_names(), self.__dict__['values']))
        return namespace_cls(**d)

    # noinspection PyArgumentList
    def namedtuple(self):
        row_nt = namedtuple('Row', self.header_names())
        return row_nt(*self.__dict__['values'])

    # noinspection PyProtectedMember,PyUnusedLocal
    def join_values(self, row_b, on_columns=None):
        """ copies all values where header names are shared with row_b
        :type row_b: flux_row_cls
        :type on_columns: shared column names
        """
        # region {closure function}
        def intersecting_names(_names_):
            if not _names_:
                _names_ = headers_a.keys() & headers_b.keys()
                if not _names_:
                    raise ValueError('no intersecting column names')

            return _names_
        # endregion

        if not isinstance(row_b, flux_row_cls):
            raise TypeError('row expected to be flux_row_cls')

        headers_a = self.__dict__['_headers']
        headers_b = row_b._headers

        values_a  = self.__dict__['values']
        values_b  = row_b.values

        names = intersecting_names(on_columns)

        for name in names:
            i_b = headers_b[name]
            i_a = headers_a[name]

            values_a[i_a] = values_b[i_b]

    def __getattr__(self, name):
        """  eg:
             o = row.column
        """
        try:
            i = self.__dict__['_headers'].get(name, name)
            return self.__dict__['values'][i]
        except (TypeError, IndexError):
            self.__raise_attribute_error(name, self.headers)

    def __setattr__(self, name, value):
        """ eg:
            row.column = o
        """
        try:
            i = self.__dict__['_headers'].get(name, name)
            self.__dict__['values'][i] = value
        except (TypeError, IndexError) as e:

            if name in self.__dict__:
                self.__dict__[name] = value
            elif isinstance(self.__dict__['values'], tuple):
                raise e
            else:
                self.__raise_attribute_error(name, self.headers)

    def __getitem__(self, name):
        """ eg:
            o = row['column']
        """
        try:
            i = self.__dict__['_headers'].get(name, name)
            return self.__dict__['values'][i]
        except (TypeError, IndexError):
            if isinstance(name, slice):
                return self.__dict__['values'][name]

            self.__raise_attribute_error(name, self.headers)

    def __setitem__(self, name, value):
        """ eg:
            row['column'] = o
        """
        try:
            i = self.__dict__['_headers'].get(name, name)
            self.__dict__['values'][i] = value
        except (TypeError, IndexError) as e:

            if isinstance(self.__dict__['values'], tuple):
                raise e
            elif isinstance(name, slice):
                self.__dict__['values'][name] = value
            elif name in self.__dict__:
                self.__dict__[name] = value
            else:
                self.__raise_attribute_error(name, self.headers)

    def __len__(self):
        return len(self.__dict__['values'])

    def __bool__(self):
        return bool(self.__dict__['values'])

    def __iter__(self):
        return iter(self.__dict__['values'])

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(self.__dict__['_headers']) + hash(tuple(self.__dict__['values']))

    def __repr__(self):
        if 'address' in self.__dict__:
            row_label = format_header_lite(self.__dict__['address'])
            row_label = ' {}   '.format(row_label)
        elif 'r_i' in self.__dict__:
            row_label = format_integer(self.__dict__['r_i'])
            row_label = format_header_lite(row_label)
            row_label = ' {}   '.format(row_label)
        else:
            row_label = ''

        if self.is_jagged():
            jagged_label = len(self.__dict__['values']) - len(self.__dict__['_headers'])
            jagged_label = format_header_lite('{:+}'.format(jagged_label))
            jagged_label = ' 🗲jagged {}🗲  '.format(jagged_label)
        else:
            jagged_label = ''

        if self.is_header_row():
            values = ', '.join(str(n) for n in self.header_names())
            values = format_header(values)
        else:
            values = (repr(self.__dict__['values']).replace('"', '')
                                                   .replace("'", ''))

        return '{}{}{}'.format(row_label, jagged_label, values)

    @staticmethod
    def __raise_attribute_error(invalid, headers):
        indices = [str(i) for i in headers.values()]
        _nf_ = max(len(n) for n in indices)
        _nf_ = '{: <%s}' % str(_nf_)
        indices = [_nf_.format(n) for n in indices]

        s = ["{}: '{}'".format(i, n) for i, n in zip(indices, headers.keys())]
        s = '\n\t'.join(s)

        s = ("'{}' column name does not exist in columns: "
             "\n\t{}".format(invalid, s))

        raise AttributeError(s) from None





from .text import print_runtime
from .text import print_performance
from .text import styled

from .dates import is_date
from .dates import to_datetime
from .dates import parse_timedelta
from .dates import parse_seconds

from .filesystem import read_file
from .filesystem import write_file
from .filesystem import parse_path
from .filesystem import traverse_dir

from .iter import transpose


__all__ = ['print_runtime',
           'print_performance',
           'styled',

           'is_date',
           'to_datetime',
           'parse_timedelta',
           'parse_seconds',

           'read_file',
           'write_file',
           'parse_path',
           'traverse_dir',

           'transpose']




import os
import sys

''' 
config: 
    default settings for vengeance.util functions
    eg:
        [console]
        effect = bold
        enable_ansi_escape = True
        formatter = {vengeance_prefix}@{name}: {elapsed}

loads_excel_module: 
    determines if excel_com module should be loaded in vengeance.__init__ 
    if environment is expected to support Windows COM interface, load vengeance.excel_com module
    
    vengeance.excel_com functions:
        vengeance.open_workbook
        vengeance.close_workbook
        vengeance.excel_levity_cls
        etc ...

ordereddict:
    starting at python 3.6, the built-in dict is both insertion-ordered and compact, 
    using about half the memory of collections.OrderedDict 

'''
config          = {}
__config_already_read__ = False

python_version      = sys.version_info
is_windows_os       = (os.name == 'nt') or (sys.platform == 'win32')
is_utf_console      = ('utf' in sys.stdout.encoding.lower())
is_tty_console      = False
is_pypy_interpreter = ('__pypy__' in iter(sys.builtin_module_names))
loads_excel_module  = (not is_pypy_interpreter) and is_windows_os

ordereddict             = dict
dateutil_installed      = False
ultrajson_installed     = False
numpy_installed         = False
line_profiler_installed = False


if python_version < (3, 6):
    from collections import OrderedDict
    ordereddict = OrderedDict

if python_version < (3, 9):
    try:
        import ujson
        ultrajson_installed = True
    except ImportError:
        pass

try:
    import dateutil
    dateutil_installed = True
except ImportError:
    pass

try:
    import numpy
    numpy_installed = True
except ImportError:
    pass

try:
    import line_profiler
    line_profiler_installed = True
except ImportError:
    pass


def enable_ansi_escape_in_console():
    """ attempt to turn on ansi escape colors in console

    ansi escape: \x1b, \033, <0x1b>, 
    how to effectively check if console supports ansi escapes?
        ansi escape shows up as '←' in python.exe console

    STD_OUTPUT_HANDLE = -11
    DWMODE            = 7
    """
    global is_utf_console
    global is_tty_console

    if not is_windows_os:
        return

    try:
        import ctypes
        kernel32 = ctypes.windll.kernel32
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)

        is_utf_console = True
        is_tty_console = False
    except:
        pass


def read_config_file():
    global config
    global __config_already_read__

    from configparser import ConfigParser

    __config_already_read__ = True

    config_path = __locate_config_path()
    if config_path is None:
        return

    cp = ConfigParser()
    cp.read(config_path)

    if cp.has_section('console'):
        cp_section = cp['console']

        config.update({'color':              cp_section.get('color'),
                       'effect':             cp_section.get('effect'),
                       'formatter':          cp_section.get('formatter'),
                       'enable_ansi_escape': cp_section.getboolean('enable_ansi_escape')})

        if config['enable_ansi_escape']:
            enable_ansi_escape_in_console()

    # if cp.has_section('filesystem'):
    #     cp_section = cp['filesystem']
    #     config.update({'encoding': cp_section.get('encoding')})


def __locate_config_path():
    import site

    config_paths = [site.getsitepackages()[1]  + '\\vengeance\\config.ini',
                    os.environ['localappdata'] + '\\Temp\\vengeance\\config.ini']
    if not is_windows_os:
        del config_paths[-1]

    for config_path in config_paths:
        if os.path.exists(config_path):
            return config_path

    return None


if __config_already_read__ is False:
    read_config_file()

# is_tty_console          = (hasattr(sys.stdout, 'isatty') and sys.stdout.isatty())
# is_pypy_interpreter     = True
# is_windows_os           = True
# dateutil_installed      = False
# ultrajson_installed     = False
# numpy_installed         = False
# line_profiler_installed = False
# loads_excel_module      = False






from collections import Iterable
from collections import ItemsView
from collections import KeysView
from collections import ValuesView
from collections import namedtuple

from itertools import chain
from typing import Generator
from typing import Union
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

from ..conditional import ordereddict
from .classes.tree_cls import tree_cls
from .classes.namespace_cls import namespace_cls


class IterationDepthError(TypeError):
    pass


class ColumnNameError(ValueError):
    pass


def iteration_depth(values, first_element_only=True):
    """
    this function is heavily utilized to correctly un-nest function arguments
    in flux_cls methods

    eg:
        0 = iteration_depth('abc')
        1 = iteration_depth(['abc'])
        2 = iteration_depth([[2], [2]])

        items = [1, [2, 2, [3, 3, 3, [4, [5, 5, 5, 5], 4]]]]
        1 = iteration_depth(items, first_element_only=True)
        5 = iteration_depth(items, first_element_only=False)
    """
    if is_exhaustable(values):
        raise TypeError('cannot evaluate an exhaustable iterator')

    if isinstance(values, dict):
        values = tuple(values.values())

    if not is_collection(values):
        return 0
    if len(values) == 0:
        return 1
    if first_element_only:
        return 1 + iteration_depth(values[0], first_element_only)
    else:
        return 1 + max([iteration_depth(v, first_element_only) for v in values])


def modify_iteration_depth(values,
                           depth=None,
                           depth_offset=None,
                           first_element_only=True):
    """
    this function is heavily utilized to correctly un-nest function arguments
    in flux_cls methods

    eg:
        'a'              = modify_iteration_depth([[['a']]], depth=0)
        ['a', 'b']       = modify_iteration_depth([[['a', 'b']]], depth=0)
        [['a']]          = modify_iteration_depth('a', depth=2)
        [[[['a', 'b']]]] = modify_iteration_depth(['a', 'b'], depth=4)

        required behaviors:
            must not reduce depth of any dictionary keys or values
                {'a': None} = modify_iteration_depth([{'a': None}], depth=0)

        undecided behaviors:
            how to deal with mixed iteration depths?
                [['ab', ['cd']]]
                should this return ['ab', ['cd']] or ['ab', 'cd']?

            # has_multiple_depths = set(iteration_depth(_v_, first_element_only=False) for _v_ in v)

    """

    if depth is None and depth_offset is None:
        raise ValueError('conflicting values for depth and depth_offset')

    if isinstance(depth, int) and isinstance(depth_offset, int):
        raise ValueError('conflicting values for depth and depth_offset')

    if depth_offset is None:
        value_depth  = iteration_depth(values, first_element_only=first_element_only)
        depth_offset = depth - value_depth

    if depth_offset < 0:
        for _ in range(abs(depth_offset)):

            if first_element_only:
                if is_descendable(values): values = values[0]
                else:                      break
            else:
                values = list(chain(*values))
                # values = list(chain.from_iterable(values))

    elif depth_offset > 0:
        for _ in range(depth_offset):
            values = [values]

    return values


def standardize_variable_arity_values(values,
                                      depth=None,
                                      depth_offset=None):
    """
    this function is heavily utilized to correctly un-nest function arguments
    in flux_cls methods

    eg flux_cls:
        def append_columns(self, *names):
            names = standardize_variable_arity_values(names, depth=1)
    """
    # convert iterators, even if they are nested
    if is_descendable(values) and is_exhaustable(values[0]):
        values = iterator_to_collection(values[0])
    else:
        values = iterator_to_collection(values)

    values = modify_iteration_depth(values, depth, depth_offset,
                                    first_element_only=True)

    return values


def are_indices_contiguous(indices):
    if len(indices) == 1:
        return False

    for i_2, i_1 in zip(indices[1:], indices):
        if i_2 - i_1 != 1:
            return False

    return True


def is_collection(o):
    """ determine if value is an iterable object or data structure
    function used mainly to distinguish data structures from other 
    iterables

    eg:
        False = is_collection('bleh')
        True  = is_collection(['bleh'])

        False = is_collection(range(3))
        True  = is_collection(list(range(3)))
    """
    if isinstance(o, (str, bytes)):
        return False

    if isinstance(o, range):
        return False

    return isinstance(o, Iterable)


# noinspection PyStatementEffect,PyBroadException
def is_subscriptable(o):
    """
    from typing import Sequence
    isinstance(o, Sequence)
    """
    try:
        o[0]
        return True
    except Exception:
        return False


def is_exhaustable(o):
    """
    return (hasattr(o, '__next__') or
            isinstance(o, range))
    """
    return hasattr(o, '__next__')


def is_dictview(o):
    return (isinstance(o, KeysView) or
            isinstance(o, ValuesView) or
            isinstance(o, ItemsView))


def is_vengeance_class(o):
    base_cls_names      = set(base_class_names(o))
    vengeance_cls_names = {'flux_cls',
                           'excel_levity_cls'}

    return bool(base_cls_names & vengeance_cls_names)


# def is_namedtuple_class(o):
#     return (isinstance(o, tuple) and
#             type(o) is not tuple)


def is_descendable(o):
    return (is_collection(o) and
            is_subscriptable(o) and
            len(o) == 1)


def iterator_to_collection(o):
    if is_vengeance_class(o):
        return list(o.rows())

    if isinstance(o, (range, set)) or \
       is_exhaustable(o) or \
       is_dictview(o):

        return list(o)

    return o


def base_class_names(o):
    return [b.__name__ for b in o.__class__.mro()]


def transpose(m, astype=None) -> Generator[Union[List, Tuple], None, None]:
    m = iterator_to_collection(m)
    n = iteration_depth(m, first_element_only=True)

    if n == 0:
        raise IterationDepthError('matrix must have at least 1 iterable dimension')

    if astype is None:
        if n == 1: astype = type(m)
        else:      astype = type(m[0])

    if astype is list:
        if n == 1: t = ([row] for row in m)
        else:      t = (list(row) for row in zip(*m))
    else:
        if n == 1: t = ((row,) for row in m)
        else:      t = zip(*m)

    return t


# noinspection PyProtectedMember
def to_namespaces(o):
    """ recursively convert values to namespace """

    # region {closure functions}
    # noinspection PyProtectedMember
    def traverse(k, v):
        if isinstance(v, dict):
            return to_namespaces(v)
        elif hasattr(v, '_asdict'):
            return to_namespaces(v._asdict())
        elif is_collection(v):
            return [traverse(k, _) for _ in v]
        else:
            return v
    # endregion

    if hasattr(o, '_asdict'):
        o = o._asdict()

    if isinstance(o, dict):
        d = {k: traverse(k, v) for k, v in o.items()}
        return namespace_cls(**d)
    elif is_collection(o):
        return [traverse(None, v) for v in o]
    else:
        return o


# noinspection PyArgumentList,PyProtectedMember
def to_namedtuples(o):
    """ recursively convert values to namedtuples """

    # region {closure functions}
    # noinspection PyProtectedMember
    def traverse(v):
        if isinstance(v, dict):
            return to_namedtuples(v)
        elif hasattr(v, '_asdict'):
            return to_namedtuples(v._asdict())
        elif is_collection(v):
            return [traverse(_) for _ in v]
        else:
            return v
    # endregion

    if hasattr(o, '_asdict'):
        o = o._asdict()

    if isinstance(o, dict):
        nt = namedtuple('nt', o.keys())
        return nt(*[traverse(v) for v in o.values()])
    elif is_collection(o):
        return [traverse(v) for v in o]
    else:
        return o


def to_grouped_dict(flat_dict: Dict) -> Dict[Any, Dict]:
    """ re-map keys as tuples to a nested structure

    eg:
        # "flat" keys
        d = {('a₁', 'b₁', 'c₁', 'd₁'): 'v_1',
             ('a₁', 'b₂', 'c₁', 'd₂'): 'v_2',
             ('a₂', 'b₁', 'c₁', 'd₁'): 'v_4',
             ('a₂', 'b₁', 'c₁', 'd₂'): 'v_3'}

        {'a₁': {'b₁': {'c₁': {'d₁': 'v_1'}},
                'b₂': {'c₁': {'d₂': 'v_2'}}},
         'a₂': {'b₁': {'c₁': {'d₁': 'v_4',
                              'd₂': 'v_3'}}}} = to_grouped_dict(d)
    """
    groups = tree_cls(flat_dict).traverse()

    return groups


def map_values_to_enum(sequence, start=0) -> Dict[Union[str, bytes], int]:
    """ :return {unique_key: i: int} for all items in sequence

    values are modified before they are added as keys:
        * values coerced to string (if not bytes)
        * non-unique keys are appended with '_dup_{num}' suffix

    eg
        {'a':       0,
         'b':       1,
         'c':       2,
         'b_dup_2': 3,
         'b_dup_3': 4} = map_values_to_enum(['a', 'b', 'c', 'b', 'b'])
    """
    duplicates = {}
    values_to_indices = ordereddict()

    for i, v in enumerate(sequence, start):

        is_bytes = isinstance(v, bytes)
        if is_bytes:
            v_str = v.decode()
        else:
            v_str = str(v)

        is_duplicate = (v_str in duplicates)

        if not is_duplicate:
            duplicates[v_str] = 1
        else:
            duplicates[v_str] += 1

        if is_duplicate:
            v_str = '{}_dup_{}'.format(v_str, duplicates[v_str])

        if is_bytes:
            v_str = v_str.encode()

        values_to_indices[v_str] = i

    return values_to_indices


def is_header_row(values, headers):
    """ determine if underlying values match headers.keys

    checking headers.keys() == values will not always work, since map_values_to_enum()
    was used to modify names into more suitable header keys
    """
    if not headers:
        return False

    value_names  = map_values_to_enum(values).keys()
    header_names = set(headers)

    return value_names == header_names





import os
import setuptools
from setuptools import setup

"""
Publishing (Perfect) Python Packages on PyPi
https://youtu.be/GIF3LaRqgXo

?
https://hynek.me/articles/conditional-python-dependencies/

pip install vengeance[comtypes]
pip install vengeance[pypiwin32]
pip install vengeance[python-dateutil]
pip install vengeance[numpy]
pip install vengeance[comtypes,pypiwin32,python-dateutil,numpy]
"""


__version__ = '1.1.18'
__release__ = '$release 55'
long_description = ('https://github.com/michael-ross-ven/vengeance/blob/master/README.md'
                    '\n\n(specialize this for pypi.org later)')

is_windows_os = (os.name == 'nt')

if is_windows_os:
    install_requires = ['comtypes',
                        'pypiwin32']
else:
    install_requires = []

extras_require = {":python_version>='3.0'": ['python-dateutil',
                                             'numpy']}



def __move_win32com_gencache_folder():
    """
    move win32com gen_py cache files from temp to site-packages folder
    helps prevent win32com EnsureDispatch() call rejection due to corrupted COM files
    """
    if not is_windows_os:
        return
        
    try:
        import shutil
        import site

        # win32com site-package installed?
        if not os.path.exists(site.getsitepackages()[1] + '\\win32com\\'):
            return

        new_gencache_folder = site.getsitepackages()[1] + '\\win32com\\gen_py\\'
        old_gencache_folder = os.environ['userprofile'] + '\\AppData\\Local\\Temp\\gen_py\\'

        if not os.path.exists(new_gencache_folder):
            os.makedirs(new_gencache_folder)
            if os.path.exists(old_gencache_folder):
                shutil.rmtree(old_gencache_folder)

    except Exception:
        pass


if __name__ == '__main__':
    setup(name='vengeance',
          version=__version__,
          description='Library focusing on row-major organization of tabular data and control over '
                      'the Excel application',
          long_description=long_description,
          url='https://github.com/michael-ross-ven/vengeance',
          author='Michael Ross',
          author_email='',
          license='MIT',
          install_requires=install_requires,
          extras_require=extras_require,
          packages=setuptools.find_packages(),
          classifiers=[
              "Programming Language :: Python :: 3",
              "License :: OSI Approved :: MIT License"
            ]

          )

    if is_windows_os:
        __move_win32com_gencache_folder()

